from io import BytesIO
from typing import List, Optional, Tuple, Callable
from asyncio import get_event_loop, AbstractEventLoop
from aiohttp import ClientSession

from .utils import executor, RequestError
from .sync_docs import SyncScraper

class AsyncScraper(SyncScraper):

    def __init__(self,
        loop : Optional[AbstractEventLoop] = None, 
        session: Optional[ClientSession]   = None,
    ):
        super().__init__()

        if loop:
            self.loop = loop
        else:
            self.loop = get_event_loop()

        if session:
            self.session = session
        
    @executor()
    def _fuzzy_finder(self, *args, **kwargs):
        return super()._fuzzy_finder(*args, **kwargs)

    @executor()
    def _parse_bytes(self, data: bytes):
        yield from super()._parse_bytes(data)

    @executor()
    def _split_line(self, url: str, data: str) -> None:
        return super()._split_line(url, data)

    async def search(self, query: str, *, page: str):

        if not hasattr(self, "session"):
            self.session = ClientSession()

        if not page.endswith("/"):
            page += "/"

        if not self.cache.get(page):
            async with self.session.get(page + "objects.inv") as r:

                if r.ok:
                    data = BytesIO(await r.read())

                    for _ in range(4):
                        data.readline()

                    data = "".join(await self._parse_bytes(data))
                    await self._split_line(page, data)

                elif r.status == 404:
                    raise TypeError("Invalid documentation url, url provided does not have an objects.inv")
                else:
                    raise RequestError(f"{r.status} {r.reason}")

        data = await self._fuzzy_finder(
            query = query, 
            collection = list(self.cache[page].items()), 
        )
        return data


from .async_docs import AsyncScraper
from .sync_docs  import SyncScraper
from .utils import executor

__version__ = "0.1.0"
__author__  = "Tom the Bomb"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Tom the Bomb"



import requests
import os
from io import BytesIO
from typing import List, Optional, Tuple, Callable, Coroutine, Any

import zlib
import re

from .utils import RequestError

class SyncScraper:

    def __init__(self):
        self._line_regex = re.compile(r'(?x)(.+?)\s+(\S*:\S*)\s+(-?\d+)\s+(\S+)\s+(.*)')
        self.cache = {}

    def _fuzzy_finder(self, query: str, collection: List[Tuple[str, str]]):
        suggestions = []
        pat   = '.*?'.join(map(re.escape, query))
        regex = re.compile(pat, flags = re.IGNORECASE)

        for k, v in collection:
            out = regex.search(k)
            if out:
                suggestions.append((len(out.group()), out.start(), (k,v)))

        return [z for _, _, z in sorted(
            suggestions, 
            key = lambda tup: (tup[0], tup[1], tup[2][0])
        )]

    def _parse_bytes(self, data: bytes):
        decompressor = zlib.decompressobj()
        while True:
            chunk = data.read(16384)
            if len(chunk) == 0:
                break
            line = decompressor.decompress(chunk)
            yield line.decode("utf-8")

    def _split_line(self, url: str, data: str) -> None:
        
        self.cache[url] = {}
        for line in data.split("\n"):
            match = self._line_regex.match(line.rstrip())
            if not match:
                continue

            name, __, __, path, display = match.groups()

            path = path.strip("$") + name if path.endswith("$") else path
            key  = name if display == '-' else display

            self.cache[url][key] = os.path.join(url, path)

        return self.cache[url]

    def search(self, query: str, *, page: str):

        if not page.endswith("/"):
            page += "/"

        if not self.cache.get(page):
            resp = requests.get(page + "objects.inv")
            
            if resp.ok:
                data = BytesIO(resp.content)

                for _ in range(4):
                    data.readline()

                data = "".join(self._parse_bytes(data))
                self._split_line(page, data)

            elif resp.status_code == 404:
                raise TypeError("Invalid documentation url, url provided does not have an objects.inv")
            else:
                raise RequestError(f"{resp.status} {resp.reason}")

        data = self._fuzzy_finder(
            query = query, 
            collection = list(self.cache[page].items()), 
        )
        return data


import functools
from asyncio import get_running_loop

def executor(loop = None, executor = None):

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            partial = functools.partial(func, *args, **kwargs)
            loop    = get_running_loop()
            return loop.run_in_executor(executor, partial)

        return wrapper
    return decorator

class RequestError(Exception):
    pass


from setuptools import setup

import re

with open("doc_search/__init__.py") as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

setup(
    name         = "doc-search", 
    author       = "Tom-the-Bomb", 
    version      = version, 
    description  = "A package for searching documentation built with sphinx",
    long_description              = open("README.md").read(),
    long_description_content_type = "text/markdown",
    license      = "MIT",
    url          = "https://github.com/Tom-the-Bomb/doc-search",
    project_urls = {
        "Repository"   : "https://github.com/Tom-the-Bomb/doc-search",
        "Issue tracker": "https://github.com/Tom-the-Bomb/doc-search/issues",
    },
    classifiers  = [
        "Intended Audience :: Developers",
        'Programming Language :: Python',
        'Natural Language :: English',
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
    ],
    include_package_data = True,
    packages             = ['doc_search'],
    install_requires     = ['aiohttp', 'requests'],
    zip_safe        = True,
    python_requires = '>=3.7'
)